/* -*-c-*-
 *
 * Tunnel interface for 4.4BSD-derived systems
 *
 * (c) 2001 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Trivial IP Encryption (TrIPE).
 *
 * TrIPE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * TrIPE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TrIPE; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#define TUN_INTERNALS

#include "tripe.h"

/*----- Main code ---------------------------------------------------------*/

#ifdef TUN_BSD

struct tunnel {
  const tunnel_ops *ops;		/* Pointer to operations */
  sel_file f;				/* Selector for tunnel device */
  struct peer *p;			/* Pointer to my peer */
};

/* --- @t_read@ --- *
 *
 * Arguments:	@int fd@ = file descriptor to read
 *		@unsigned mode@ = what's happened
 *		@void *v@ = pointer to tunnel block
 *
 * Returns:	---
 *
 * Use:		Reads data from the tunnel.
 */

static void t_read(int fd, unsigned mode, void *v)
{
  tunnel *t = v;
  ssize_t n;
  buf b;

  n = read(fd, buf_i, sizeof(buf_i));
  if (n < 0) {
    a_warn("TUN", "%s", p_ifname(t->p), "read-error", "?ERRNO", A_END);
    return;
  }
  IF_TRACING(T_TUNNEL, {
    trace(T_TUNNEL, "tun-bsd: packet arrived");
    trace_block(T_PACKET, "tun-bsd: packet contents", buf_i, n);
  })
  buf_init(&b, buf_i, n);
  p_tun(t->p, &b);
}

/* --- @t_init@ --- *
 *
 * Arguments:	---
 *
 * Returns:	---
 *
 * Use:		Initializes the tunneling system.  Maybe this will require
 *		opening file descriptors or something.
 */

static void t_init(void) { return; }

/* --- @t_open@ --- *
 *
 * Arguments:	@char **ifn@ = where to put the interface name
 *
 * Returns:	A file descriptor, or @-1@ on failure.
 *
 * Use:		Opens a tunnel device.  This will run with root privileges
 *		even if the rest of the server has dropped them.
 */

static int t_open(char **ifn)
{
  int fd;
  unsigned n;
  char buf[16];

  n = 0;
  for (;;) {
    sprintf(buf, "/dev/tun%u", n);
    if ((fd = open(buf, O_RDWR)) >= 0)
      break;
    switch (errno) {
      case EBUSY:
       T( trace(T_TUNNEL, "tunnel device %u busy: skipping", n); )
       break;
      case ENOENT:
       a_warn("TUN", "-", "bsd", "no-tunnel-devices", A_END);
       return (-1);
      default:
       a_warn("TUN", "-", "open-error", "%s", buf, "?ERRNO", A_END);
       break;
    }
    n++;
  }
  return (fd);
}

/* --- @t_create@ --- *
 *
 * Arguments:	@peer *p@ = pointer to peer block
 *		@int fd@ = file descriptor of tunnel device
 *		@char **ifn@ = where to put the interface name
 *
 * Returns:	A tunnel block if it worked, or null on failure.
 *
 * Use:		Initializes a new tunnel.
 */

static tunnel *t_create(peer *p, int fd, char **ifn)
{
  tunnel *t;

  fdflags(fd, O_NONBLOCK, O_NONBLOCK, FD_CLOEXEC, FD_CLOEXEC);
  t = CREATE(tunnel);
  t->ops = &tun_bsd;
  t->p = p;
  sel_initfile(&sel, &t->f, fd, SEL_READ, t_read, t);
  sel_addfile(&t->f);
  return (t);
}

/* --- @t_inject@ --- *
 *
 * Arguments:	@tunnel *t@ = pointer to tunnel block
 *		@buf *b@ = buffer to send
 *
 * Returns:	---
 *
 * Use:		Injects a packet into the local network stack.
 */

static void t_inject(tunnel *t, buf *b)
{
  IF_TRACING(T_TUNNEL, {
    trace(T_TUNNEL, "tun-bsd: inject decrypted packet");
    trace_block(T_PACKET, "tun-bsd: packet contents", BBASE(b), BLEN(b));
  })
  write(t->f.fd, BBASE(b), BLEN(b));
}

/* --- @t_destroy@ --- *
 *
 * Arguments:	@tunnel *t@ = pointer to tunnel block
 *
 * Returns:	---
 *
 * Use:		Destroys a tunnel.
 */

static void t_destroy(tunnel *t)
  { sel_rmfile(&t->f); close(t->f.fd); DESTROY(t); }

const tunnel_ops tun_bsd = {
  "bsd",
  t_init,
  t_open,
  t_create,
  0,
  t_inject,
  t_destroy
};

#endif

/*----- That's all, folks -------------------------------------------------*/
